# $Id: table.py 30 2006-02-20 23:57:51Z psanchez $
#
# Copyright 2005, Pedro I. Sanchez (psanchez@colcan.biz)
#
#     This computer program is released under the
#         GNU GENERAL PUBLIC LICENSE Version 2
#              (See the file COPYING)
#

from random import randrange

OPER_ADDITION       = 1
OPER_SUBSTRACTION   = 2
OPER_MULTIPLICATION = 3
OPER_DIVISION       = 4
OPER_DEFAULT        = OPER_ADDITION

OR_HORIZONTAL = 0
OR_VERTICAL   = 1
OR_DEFAULT    = OR_VERTICAL

operators = {OPER_ADDITION: '+',
             OPER_SUBSTRACTION: '-',
             OPER_MULTIPLICATION: 'x',
             OPER_DIVISION: '/'}

# these are the default values
MINL = 2
MAXL = 5
MINR = 2
MAXR = 12
LINES= 35
COLS = 4

# this is the expected number of characters per line
# If works well with the default fonts and could be
# embarrassingly bad for others.
LINE_SIZE = 66

class Layout(object):
    """ Horizontal formatting for all four operations:

                (student sheet)      (answer sheet)

                125 x 33 = ____      125 x 33 = 4125

                <--lcell--> x <--rcell--> = <--acell-->

        Vertical formatting for '+', '-', and 'x' operations:

                (student sheet)      (answer sheet)

                   125                  125
                 x  33                x  33
                 -----                -----
                                       4125

                  <--lcell-->
                x <--rcell-->
                -------------
                <---acell--->

        Vertical formatting for the division operation (North America):

                (student sheet)      (answer sheet)

                     _____                _33__
                125 ( 4125           125 ( 4125

                             ____________
                <--lcell--> ( <--rcell-->
                            <---acell--->

        Vertical formatting for the division operation (Latin America):

                (student sheet)      (answer sheet)

                4125 | 125           4125 | 125
                     -----                -----
                                            33

                <--lcell--> | <--rcell-->
                            -------------
                            <---acell--->
            
    """
    def __init__(self):
        self.lines = LINES
        self.cols  = COLS
        self.lsize = LINE_SIZE
        self.minl  = MINL
        self.maxl  = MAXL
        self.minr  = MINR
        self.maxr  = MAXR
        self.acv = 0
        self.nmax = 0
        self.pmax = 0
        self.lcell = 0
        self.rcell = 0
        self.acell = 0
        self.vspace = 0
        self.height = 0
        self.line  = ''
        self.linechar = '-'
        self.orientation = OR_DEFAULT
        self.operator    = OPER_DEFAULT
        self._format()
        self.setACV(0)
            
    def _setNumCellsWidth(self):
        """ Calculates lcell and rcell according to the maximum
            values that each math operation can generate.
        """
        self.lcell = self.nmax
        self.rcell = self.nmax
        if self.operator == OPER_DIVISION:
            if self.orientation == OR_HORIZONTAL:
                self.lcell = self.pmax
            elif self.orientation == OR_VERTICAL:
                self.rcell = self.pmax

    def _setAnsCellWidth(self):
        """ Calculates the length of the answer cell, acell,
            according to the type of requested operation.
        """
        if self.orientation == OR_HORIZONTAL:
            # in all but the 'x' case we take the maximum
            # number plus one to accomodate carry overs
            self.acell = self.nmax+1
            if self.operator == OPER_MULTIPLICATION:
                self.acell = self.pmax
        else:
            # this is the vertical case.
            # in the '+' and '-' cases we take the maximum
            # number plus two (one space+operator)
            self.acell = self.nmax+2
            if self.operator == OPER_MULTIPLICATION or \
               self.operator == OPER_DIVISION:
                self.acell = self.pmax+2

    def _setAnsLine(self):
        """ Creates a 'line' as long as the answer cell.
        """
        self.linechar = '-'
        if self.orientation == OR_HORIZONTAL or \
           (self.operator == OPER_DIVISION and \
            self.orientation == OR_VERTICAL):
            self.linechar = '_'
        self.line = ''
        n = self.acell
        if self.operator == OPER_DIVISION and \
           self.orientation == OR_VERTICAL:
            n -= 1
        self.line += self.linechar*n

    def _format(self):
        self.nmax = len(str(max(self.maxl, self.maxr)))
        self.pmax = len(str(self.maxl*self.maxr))
        self.height = 1
        # leave at least 1 empty lines between rows
        # of operations
        self.vspace = 1
        # in vertical mode we need room to do the
        # operations
        if self.orientation == OR_VERTICAL:
            self.height = 4
            if self.operator == OPER_MULTIPLICATION:
                self.vspace = self.nmax
            elif self.operator == OPER_DIVISION:
                self.vspace = self.pmax
                self.height = 2
        self.height += self.vspace
        self._setNumCellsWidth()
        self._setAnsCellWidth()
        self._setAnsLine()

    def getLinesCols(self):
        return self.lines, self.cols

    def getCellHeight(self):
        return self.height

    def getMinMax(self):
        return self.minl, \
               self.maxl, \
               self.minr, \
               self.maxr

    def getACV(self):
        return self.acv

    def getOperator(self):
        return self.operator

    def setACV(self, data):
        acv = data
        if data == 0: # reset to default values
            acv = 0
            if self.operator == OPER_ADDITION:
                acv = self.maxl+self.maxr
            elif self.operator == OPER_MULTIPLICATION:
                acv = self.maxl*self.maxr
        self.acv = acv

    def setLinesCols(self, lines, cols):
        if lines < 1 or cols < 1:
            raise ValueError
        self.lines = lines
        self.cols  = cols

    def setMinMax(self, minl, maxl, minr, maxr):
        if maxl < minl or maxr < minr:
            raise ValueError
        self.minl = minl
        self.maxl = maxl
        self.minr = minr
        self.maxr = maxr
        self._format()

    def setOperator(self, data):
        if data != OPER_ADDITION and \
           data != OPER_SUBSTRACTION and \
           data != OPER_MULTIPLICATION and \
           data != OPER_DIVISION:
            raise ValueError
        self.operator = data
        self._format()

    def setOrientation(self, data):
        if data != OR_HORIZONTAL and \
           data != OR_VERTICAL:
            raise ValueError
        self.orientation = data
        self._format()

def random_numbers(layout):
    minl, maxl, minr, maxr = layout.getMinMax()
    acv = layout.getACV()
    operator = layout.getOperator()
    # loop until we get a pair of numbers that
    # satisfies the ceiling condition of the answer.
    l = randrange(minl, maxl+1)
    r = randrange(minr, maxr+1)
    if operator == OPER_ADDITION:
        while l+r > acv:
            l = randrange(minl, maxl+1)
            r = randrange(minr, maxr+1)
    elif operator == OPER_MULTIPLICATION:
        while l*r > acv:
            l = randrange(minl, maxl+1)
            r = randrange(minr, maxr+1)
    # randomize numbers taking care of the
    # particular operation
    x = l
    y = r
    if operator == OPER_SUBSTRACTION:
        x = max(l,r)
        y = min(l,r)
    elif operator == OPER_DIVISION:
        if r == 0: y = randrange(1, maxr+1) 
    elif len(str(x)) < len(str(y)):
        x = r
        y = l
    return x, y

def do_answer(layout, x, y):
    operator = layout.getOperator()
    if operator == OPER_ADDITION:
        ans = str(x+y)
    elif operator == OPER_SUBSTRACTION:
        ans = str(x-y)
    elif operator == OPER_MULTIPLICATION:
        ans = str(x*y)
    elif operator == OPER_DIVISION:
        ans = str(x)
    return ans

def horz_column(layout):
    operator = layout.getOperator()
    x, y = random_numbers(layout)
    l = str(x).rjust(layout.lcell)
    if operator == OPER_DIVISION:
        l = str(x*y).rjust(layout.lcell)
    r = str(y).ljust(layout.rcell)
    ans = do_answer(layout, x, y)
    rv = '%s %s %s = ' % (l, operators[operator], r)
    yield rv+layout.line, rv+ans.ljust(layout.acell)
    return

def ver_column_div_na(layout):
    x, y = random_numbers(layout)
    # When doing division the answer is x. But we call do_answer anyway
    # to deal with non-integer division later on.
    ans = do_answer(layout, x, y)
    l = str(y).rjust(layout.lcell)
    r = str(x*y).ljust(layout.rcell)
    l2 = '%s ( %s' % (l, r)
    l1 = layout.line.rjust(len(l2))
    sans = ans.ljust(layout.rcell,'_')
    sans = '_'+sans
    l1ans = sans.rjust(len(l2))
    yield l1, l1ans
    yield l2, l2
    return

def vert_column(layout):
    operator = layout.getOperator()
    x, y = random_numbers(layout)
    ans = do_answer(layout, x, y)
    l = str(x)
    r = '%s %s' % (operators[operator], str(y))
    l1 = l.rjust(layout.acell)
    l2 = r.rjust(layout.acell)
    yield l1, l1
    yield l2, l2
    yield layout.line, layout.line
    yield ' '.rjust(layout.acell), ans.rjust(layout.acell)
    return

class Line(object):
    def __init__(self, layout):
        self.layout = layout
        self.label_cell = 0
        self.pad = 0
        self.columns = []
        if layout.orientation == OR_HORIZONTAL:
            cell = horz_column
        elif layout.orientation == OR_VERTICAL:
            if layout.operator == OPER_DIVISION:
                cell = ver_column_div_na
            else:
                cell = vert_column
        else:
            raise TypeError
        for i in range(self.layout.cols):
            self.columns.append(cell(layout))

    def _format(self, w):
        self.label_cell = len(str(self.layout.lines*self.layout.cols))+3
        col_size = self.label_cell + w
        gap = self.layout.lsize - col_size*self.layout.cols
        self.pad = ''
        for i in range(gap/self.layout.cols-1):
            self.pad += ' '

    def _render(self, n=0):
        if n < 0:
            return
        line = ''
        lineans = ''
        # format all columns but the last one on first line
        find_pad = True
        for i in range(self.layout.cols-1):
            label = '%s) ' %  (n*self.layout.cols+i+1)
            c, cans = self.columns[i].next()
            if find_pad:
                self._format(len(c))
                find_pad = False
            line += '%s%s%s' % (label.rjust(self.label_cell),
                                c, self.pad)
            lineans += '%s%s%s' % (label.rjust(self.label_cell),
                                   cans, self.pad)
        # format the last column of the first line
        label = '%s) ' %  (n*self.layout.cols+i+2)
        c, cans = self.columns[i+1].next()
        line += '%s%s\n' % (label.rjust(self.label_cell), c)
        lineans += '%s%s\n' % (label.rjust(self.label_cell), cans)
        yield line, lineans
        # format the remaining lines
        label = ''
        while True:
            # format all columns but the last one
            line = ''
            lineans = ''
            for i in range(self.layout.cols-1):
                try:
                    c, cans = self.columns[i].next()
                    line += '%s%s%s' % (label.rjust(self.label_cell),
                                        c, self.pad)
                    lineans += '%s%s%s' % (label.rjust(self.label_cell),
                                           cans, self.pad)
                except:
                    return
            # format the last column
            c, cans = self.columns[i+1].next()
            line += '%s%s\n' % (label.rjust(self.label_cell), c)
            lineans += '%s%s\n' % (label.rjust(self.label_cell), cans)
            yield line, lineans

    def render(self, n):
        g = self._render(n)
        line = ''
        lineans = ''
        vspace = '\n'*self.layout.vspace
        while True:
            try:
                li, lians = g.next()
            except:
                line += vspace
                lineans += vspace
                return line, lineans
            line += li
            lineans += lians
        
class Page(object):
    def __init__(self):
        self.layout = Layout()
        self.ans   = False
        self.anspage = ''
        self.maxlines = 46
        self.lb = chr(12)+'\n'

    def __str__(self):
        if self.ans:
            return self.anspage
        else:
            # the number of rows that fit in a page
            prows = self.maxlines/self.layout.height
            rcount = 1
            page = '\n'
            self.anspage = '\n'
            for i in range(self.layout.lines):
                if rcount >= prows:
                    page += self.lb
                    self.anspage += self.lb
                    rcount = 1
                line = Line(self.layout)
                l, lans = line.render(i)
                page += l
                self.anspage += lans
                rcount += 1
            return page

    def getLinesCols(self):
        return self.layout.getLinesCols()

    def getCellHeight(self):
        return self.layout.getCellHeight()

    def getMinMax(self):
        return self.layout.getMinMax()

    def getACV(self):
        return self.layout.getACV()

    def setACV(self, data):
        self.layout.setACV(data)

    def setMinMax(self, minl, maxl, minr, maxr):
        self.layout.setMinMax(minl, maxl, minr, maxr)

    def setLinesCols(self, lines, cols):
        self.layout.setLinesCols(lines, cols)

    def setOperator(self, data):
        self.layout.setOperator(data)
        # reset the acv when changing operations
        self.layout.setACV(0)

    def setOrientation(self, data):
        self.layout.setOrientation(data)

    def answers(self, state):
        self.ans = False
        if state: self.ans = True
